# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add resource indexes

Revision ID: daa98075b90d
Revises: c92a3e68beed
Create Date: 2021-11-26 10:26:41.883072
"""

from alembic import op
from oslo_db.sqlalchemy import utils
from oslo_log import log as logging


LOG = logging.getLogger(__name__)


# revision identifiers, used by Alembic.
revision = 'daa98075b90d'
down_revision = 'c92a3e68beed'
branch_labels = None
depends_on = None

INDEXES = (
    ('groups', 'groups_deleted_project_id_idx', ('deleted', 'project_id')),

    ('group_snapshots', 'group_snapshots_deleted_project_id_idx',
     ('deleted', 'project_id')),

    ('volumes', 'volumes_deleted_project_id_idx', ('deleted', 'project_id')),
    ('volumes', 'volumes_deleted_host_idx', ('deleted', 'host')),

    ('backups', 'backups_deleted_project_id_idx', ('deleted', 'project_id')),

    ('snapshots', 'snapshots_deleted_project_id_idx', ('deleted',
                                                       'project_id')),
)


def upgrade():
    conn = op.get_bind()
    is_mysql = conn.dialect.name == 'mysql'

    for table, idx_name, fields in INDEXES:
        # Skip creation in mysql if it already has the index
        if is_mysql and utils.index_exists(conn, table, idx_name):
            LOG.info('Skipping index %s, already exists', idx_name)
        else:
            op.create_index(idx_name, table, fields)



# Copyright 2023 toyou Corp.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest
from unittest import mock

from cinder import exception
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.volume import configuration as conf
from cinder.volume.drivers.toyou.tyds import tyds as driver

POOLS_NAME = ['pool1', 'pool2']


class TestTydsDriver(unittest.TestCase):
    @mock.patch('cinder.volume.drivers.toyou.tyds.tyds_client.TydsClient',
                autospec=True)
    def setUp(self, mock_tyds_client):
        """Set up the test case.

        - Creates a driver instance.
        - Mocks the TydsClient and its methods.
        - Initializes volumes and snapshots for testing.
        """
        super().setUp()
        self.mock_client = mock_tyds_client.return_value
        self.mock_do_request = mock.MagicMock(
            side_effect=self.mock_client.do_request)
        self.mock_client.do_request = self.mock_do_request

        self.configuration = mock.Mock(spec=conf.Configuration)
        self.configuration.tyds_pools = POOLS_NAME
        self.configuration.san_ip = "23.44.56.78"
        self.configuration.tyds_http_port = 80
        self.configuration.san_login = 'admin'
        self.configuration.san_password = 'admin'
        self.configuration.tyds_stripe_size = '4M'
        self.configuration.tyds_clone_progress_interval = 3
        self.configuration.tyds_copy_progress_interval = 3
        self.driver = driver.TYDSDriver(configuration=self.configuration)
        self.driver.do_setup(context=None)
        self.driver.check_for_setup_error()

        self.volume = fake_volume.fake_volume_obj(None)
        self.volume.host = 'host@backend#pool1'
        self.snapshot = fake_snapshot.fake_snapshot_obj(None)
        self.snapshot.volume = self.volume
        self.snapshot.volume_id = self.volume.id
        self.target_volume = fake_volume.fake_volume_obj(None)
        self.target_volume.host = 'host@backend#pool2'
        self.src_vref = self.volume

    def test_create_volume_success(self):
        """Test case for successful volume creation.

        - Sets mock return value.
        - Calls create_volume method.
        - Verifies if the create_volume method is called with correct
        arguments.
        """
        self.mock_client.create_volume.return_value = self.volume
        self.driver.create_volume(self.volume)
        self.mock_client.create_volume.assert_called_once_with(
            self.volume.name, self.volume.size * 1024, 'pool1', '4M')

    def test_create_volume_failure(self):
        """Test case for volume creation failure.

        - Sets the mock return value to simulate a failure.
        - Calls the create_volume method.
        - Verifies if the create_volume method raises the expected exception.
        """
        # Set the mock return value to simulate a failure
        self.mock_client.create_volume.side_effect = \
            exception.VolumeBackendAPIException('API error')

        # Call the create_volume method and check the result
        self.assertRaises(
            exception.VolumeBackendAPIException,
            self.driver.create_volume,
            self.volume
        )

    def test_delete_volume_success(self):
        """Test case for successful volume deletion.

        - Mocks the _get_volume_by_name method to return a volume.
        - Calls the delete_volume method.
        - Verifies if the delete_volume method is called with the correct
        volume ID.
        """
        # Mock the _get_volume_by_name method to return a volume
        self.driver._get_volume_by_name = mock.Mock(return_value={'id': '13'})

        # Call the delete_volume method
        self.driver.delete_volume(self.volume)

        # Verify if the delete_volume method is called with the correct
        # volume ID
        self.mock_client.delete_volume.assert_called_once_with('13')

    def test_delete_volume_failure(self):
        """Test case for volume deletion failure.

        - Mocks the _get_volume_by_name method to return a volume.
        - Sets the mock return value for delete_volume method to raise an
        exception.
        - Calls the delete_volume method.
        - Verifies if the delete_volume method raises the expected exception.
        """
        # Mock the _get_volume_by_name method to return a volume
        self.driver._get_volume_by_name = mock.Mock(return_value={'id': '13'})

        # Set the mock return value for delete_volume method to raise an
        # exception
        self.mock_client.delete_volume.side_effect = \
            exception.VolumeBackendAPIException('API error')

        # Call the delete_volume method and verify if it raises the expected
        # exception
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.delete_volume, self.volume)

    def test_create_snapshot_success(self):
        """Test case for successful snapshot creation.

        - Sets the mock return value for create_snapshot method.
        - Mocks the _get_volume_by_name method to return a volume.
        - Calls the create_snapshot method.
        - Verifies if the create_snapshot method is called with the correct
        arguments.
        """
        # Set the mock return value for create_snapshot method
        self.mock_client.create_snapshot.return_value = self.snapshot

        # Mock the _get_volume_by_name method to return a volume
        self.driver._get_volume_by_name = mock.Mock(return_value={'id': '13'})

        # Call the create_snapshot method
        self.driver.create_snapshot(self.snapshot)

        # Verify if the create_snapshot method is called with the correct
        # arguments
        self.mock_client.create_snapshot.assert_called_once_with(
            self.snapshot.name, '13',
            '%s/%s' % (self.volume.name, self.snapshot.name)
        )

    def test_create_snapshot_failure_with_no_volume(self):
        """Test case for snapshot creation failure when volume is not found.

        - Mocks the _get_volume_by_name method to return None.
        - Calls the create_snapshot method.
        - Verifies if the create_snapshot method is not called.
        """
        # Mock the _get_volume_by_name method to return None
        self.driver._get_volume_by_name = mock.Mock(return_value=None)

        # Call the create_snapshot method and check for exception
        self.assertRaises(driver.TYDSDriverException,
                          self.driver.create_snapshot, self.snapshot)

        # Verify if the create_snapshot method is not called
        self.mock_client.create_snapshot.assert_not_called()

    def test_create_snapshot_failure(self):
        """Test case for snapshot creation failure.

        - Mocks the _get_volume_by_name method to return a volume.
        - Sets the mock return value for create_snapshot to raise an exception.
        - Calls the create_snapshot method.
        - Verifies if the create_snapshot method is called with the correct
        arguments.
        """
        # Mock the _get_volume_by_name method to return a volume
        self.driver._get_volume_by_name = mock.Mock(return_value={'id': '13'})

        # Set the mock return value for create_snapshot to raise an exception
        self.mock_client.create_snapshot.side_effect = \
            exception.VolumeBackendAPIException('API error')

        # Call the create_snapshot method and check for exception
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_snapshot, self.snapshot)

        # Verify if the create_snapshot method is called with the correct
        # arguments
        self.mock_client.create_snapshot.assert_called_once_with(
            self.snapshot.name, '13',
            '%s/%s' % (self.volume.name, self.snapshot.name))

    def test_delete_snapshot_success(self):
        """Test case for successful snapshot deletion.

        - Mocks the _get_snapshot_by_name method to return a snapshot.
        - Calls the delete_snapshot method.
        - Verifies if the delete_snapshot method is called with the correct
        arguments.
        """
        # Mock the _get_snapshot_by_name method to return a snapshot
        self.driver._get_snapshot_by_name = mock.Mock(
            return_value={'id': 'volume_id'})

        # Call the delete_snapshot method
        self.driver.delete_snapshot(self.snapshot)

        # Verify if the delete_snapshot method is called with the correct
        # arguments
        self.mock_client.delete_snapshot.assert_called_once_with('volume_id')

    def test_delete_snapshot_failure(self):
        """Test case for snapshot deletion failure.

        - Mocks the _get_snapshot_by_name method to return a snapshot.
        - Sets the mock return value for delete_snapshot to raise an exception.
        - Calls the delete_snapshot method.
        - Verifies if the delete_snapshot method is called with the correct
        arguments.
        """
        # Mock the _get_snapshot_by_name method to return a snapshot
        self.driver._get_snapshot_by_name = mock.Mock(
            return_value={'id': 'volume_id'})

        # Set the mock return value for delete_snapshot to raise an exception
        self.mock_client.delete_snapshot.side_effect = \
            exception.VolumeBackendAPIException('API error')

        # Call the delete_snapshot method and check for exception
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.delete_snapshot,
                          self.snapshot)

        # Verify if the delete_snapshot method is called once
        self.mock_client.delete_snapshot.assert_called_once()

    @mock.patch('time.sleep')
    @mock.patch('cinder.coordination.synchronized', new=mock.MagicMock())
    def test_create_volume_from_snapshot_success(self, mock_sleep):
        """Test case for successful volume creation from snapshot.

        - Mocks the sleep function.
        - Sets the mock return values for create_volume_from_snapshot,
        _get_volume_by_name, and get_clone_progress.
        - Calls the create_volume_from_snapshot method.
        - Verifies if the create_volume_from_snapshot method is called with
        the correct arguments.
        - Verifies if the _get_volume_by_name method is called once.
        """
        # Mock the sleep function
        mock_sleep.return_value = None

        # Set the mock return values for create_volume_from_snapshot,
        # _get_volume_by_name, and get_clone_progress
        self.mock_client.create_volume_from_snapshot.return_value = self.volume
        self.driver._get_volume_by_name = mock.Mock(
            return_value={'poolName': 'pool1',
                          'sizeMB': self.volume.size * 1024})
        self.mock_client.get_clone_progress.return_value = {'progress': '100%'}

        # Call the create_volume_from_snapshot method
        self.driver.create_volume_from_snapshot(self.target_volume,
                                                self.snapshot)

        # Verify if the create_volume_from_snapshot method is called with the
        # correct arguments
        self.mock_client.create_volume_from_snapshot.assert_called_once_with(
            self.target_volume.name, 'pool2', self.snapshot.name,
            self.volume.name, 'pool1')

        # Verify if the _get_volume_by_name method is called once
        self.driver._get_volume_by_name.assert_called_once()

    def test_create_volume_from_snapshot_failure(self):
        """Test case for volume creation from snapshot failure.

        - Sets the mock return value for _get_volume_by_name to return None.
        - Calls the create_volume_from_snapshot method.
        - Verifies if the create_volume_from_snapshot method raises a
        driver.TYDSDriverException.
        """
        # Set the mock return value for _get_volume_by_name to return None
        self.driver._get_volume_by_name = mock.Mock(return_value=None)

        # Call the create_volume_from_snapshot method and check for exception
        self.assertRaises(driver.TYDSDriverException,
                          self.driver.create_volume_from_snapshot,
                          self.volume, self.snapshot)

    @mock.patch('cinder.coordination.synchronized', new=mock.MagicMock())
    def test_create_cloned_volume_success(self):
        """Test case for successful cloned volume creation.

        - Sets the mock return values for get_copy_progress, get_pools,
        get_volumes, and _get_volume_by_name.
        - Calls the create_cloned_volume method.
        - Verifies if the create_clone_volume method is called with the correct
         arguments.
        """
        # Set the mock return values for get_copy_progress, get_pools,
        # get_volumes, and _get_volume_by_name
        self.mock_client.get_copy_progress.return_value = {'progress': '100%'}
        self.driver.client.get_pools.return_value = [
            {'name': 'pool1', 'id': 'pool1_id'},
            {'name': 'pool2', 'id': 'pool2_id'}
        ]
        self.driver.client.get_volumes.return_value = [
            {'blockName': self.volume.name, 'poolName': 'pool1',
             'id': 'source_volume_id'}
        ]
        self.driver._get_volume_by_name = mock.Mock(
            return_value={'name': self.volume.name, 'id': '13'})

        # Call the create_cloned_volume method
        self.driver.create_cloned_volume(self.target_volume, self.src_vref)

        # Verify if the create_clone_volume method is called with the correct
        # arguments
        self.driver.client.create_clone_volume.assert_called_once_with(
            'pool1', self.volume.name, 'source_volume_id', 'pool2', 'pool2_id',
            self.target_volume.name
        )

    @mock.patch('cinder.coordination.synchronized', new=mock.MagicMock())
    def test_create_cloned_volume_failure(self):
        """Test case for cloned volume creation failure.

        - Sets the mock return values for get_pools and get_volumes.
        - Calls the create_cloned_volume method.
        - Verifies if the create_cloned_volume method raises a
        driver.TYDSDriverException.
        """
        # Set the mock return values for get_pools and get_volumes
        self.driver.client.get_pools.return_value = [
            {'name': 'pool1', 'id': 'pool1_id'},
            {'name': 'pool2', 'id': 'pool2_id'}
        ]
        self.driver.client.get_volumes.return_value = [
            {'blockName': self.volume.name, 'poolName': None, 'id': '14'}
        ]

        # Call the create_cloned_volume method and check for exception
        self.assertRaises(driver.TYDSDriverException,
                          self.driver.create_cloned_volume,
                          self.target_volume,
                          self.src_vref)

    def test_extend_volume_success(self):
        """Test case for successful volume extension.

        - Sets the new size.
        - Calls the extend_volume method.
        - Verifies if the extend_volume method is called with the correct
        arguments.
        """
        new_size = 10

        # Call the extend_volume method
        self.driver.extend_volume(self.volume, new_size)

        # Verify if the extend_volume method is called with the correct
        # arguments
        self.mock_client.extend_volume.assert_called_once_with(
            self.volume.name, 'pool1', new_size * 1024)

    def test_extend_volume_failure(self):
        """Test case for volume extension failure.

        - Sets the new size and error message.
        - Sets the mock side effect for extend_volume to raise an Exception.
        - Calls the extend_volume method.
        - Verifies if the extend_volume method raises the expected exception
        and the error message matches.
        - Verifies if the extend_volume method is called with the correct
        arguments.
        """
        new_size = 10

        # Set the mock side effect for extend_volume to raise an Exception
        self.mock_client.extend_volume.side_effect = \
            exception.VolumeBackendAPIException('API Error: Volume extend')

        # Call the extend_volume method and check for exception
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.extend_volume, self.volume, new_size)

        # Verify if the extend_volume method is called with the correct
        # arguments
        self.mock_client.extend_volume.assert_called_once_with(
            self.volume.name, 'pool1', new_size * 1024)

    def test_get_volume_stats(self):
        """Test case for retrieving volume statistics.

        - Sets the mock side effect for safe_get to return the appropriate
        values.
        - Sets the mock return values for get_pools and get_volumes.
        - Calls the get_volume_stats method.
        - Verifies if the get_pools and get_volumes methods are called once.
        - Verifies if the retrieved statistics match the expected statistics.
        """

        def safe_get_side_effect(param_name):
            if param_name == 'volume_backend_name':
                return 'toyou_backend'

        # Set the mock side effect for safe_get to return the appropriate
        # values
        self.configuration.safe_get.side_effect = safe_get_side_effect

        # Set the mock return values for get_pools and get_volumes
        self.mock_client.get_pools.return_value = [
            {'name': 'pool1',
             'stats': {'max_avail': '107374182400', 'stored': '53687091200'}},
            {'name': 'pool2',
             'stats': {'max_avail': '214748364800', 'stored': '107374182400'}}
        ]
        self.mock_client.get_volumes.return_value = [
            {'poolName': 'pool1', 'sizeMB': '1024'},
            {'poolName': 'pool1', 'sizeMB': '2048'},
            {'poolName': 'pool2', 'sizeMB': '3072'}
        ]

        # Call the get_volume_stats method
        stats = self.driver.get_volume_stats()

        # Verify if the get_pools and get_volumes methods are called once
        self.mock_client.get_pools.assert_called_once()
        self.mock_client.get_volumes.assert_called_once()

        # Define the expected statistics
        expected_stats = {
            'vendor_name': 'TOYOU',
            'driver_version': '1.0.0',
            'volume_backend_name': 'toyou_backend',
            'pools': [
                {
                    'pool_name': 'pool1',
                    'total_capacity_gb': 100.0,
                    'free_capacity_gb': 50.0,
                    'provisioned_capacity_gb': 3.0,
                    'thin_provisioning_support': True,
                    'QoS_support': False,
                    'consistencygroup_support': False,
                    'total_volumes': 2,
                    'multiattach': False
                },
                {
                    'pool_name': 'pool2',
                    'total_capacity_gb': 200.0,
                    'free_capacity_gb': 100.0,
                    'provisioned_capacity_gb': 3.0,
                    'thin_provisioning_support': True,
                    'QoS_support': False,
                    'consistencygroup_support': False,
                    'total_volumes': 1,
                    'multiattach': False
                }
            ],
            'storage_protocol': 'iSCSI',
        }

        # Verify if the retrieved statistics match the expected statistics
        self.assertEqual(stats, expected_stats)

    def test_get_volume_stats_pool_not_found(self):
        """Test case for retrieving volume statistics when pool not found.

        - Sets the mock return value for get_pools to an empty list.
        - Calls the get_volume_stats method.
        - Verifies if the get_pools method is called once.
        - Verifies if the get_volume_stats method raises a
        driver.TYDSDriverException.
        """
        # Set the mock return value for get_pools to an empty list
        self.mock_client.get_pools.return_value = []

        # Call the get_volume_stats method and check for exception
        self.assertRaises(driver.TYDSDriverException,
                          self.driver.get_volume_stats)

        # Verify if the get_pools method is called once
        self.mock_client.get_pools.assert_called_once()

    def test_initialize_connection_success(self):
        """Test case for successful volume initialization.

        - Sets the connector information.
        - Sets the mock return values for get_initiator_list and get_target.
        - Sets the mock return values and assertions for create_initiator_group
        , create_target, modify_target, and generate_config.
        - Calls the initialize_connection method.
        - Verifies the expected return value and method calls.
        """
        # Set the connector information
        connector = {
            'host': 'host1',
            'initiator': 'iqn.1234',
            'ip': '192.168.0.1',
            'uuid': 'uuid1'
        }

        # Set the mock return values for get_initiator_list and get_target
        self.mock_client.get_initiator_list.return_value = []
        self.mock_client.get_target.return_value = [
            {'name': 'iqn.2023-06.com.toyou:uuid1', 'ipAddr': '192.168.0.2'}]

        # Set the mock return values and assertions for create_initiator_group,
        # create_target, modify_target, and generate_config
        self.mock_client.create_initiator_group.return_value = None
        self.mock_client.create_target.return_value = None
        self.mock_client.modify_target.return_value = None
        self.mock_client.generate_config.return_value = None
        self.mock_client.get_initiator_target_connections.side_effect = [
            [],  # First call returns an empty list
            [{'target_name': 'iqn.2023-06.com.toyou:initiator-group-uuid1',
              'target_iqn': 'iqn1',
              'block': [{'name': 'volume1', 'lunid': 0}]}]
            # Second call returns a non-empty dictionary
        ]

        # Call the initialize_connection method
        result = self.driver.initialize_connection(self.volume, connector)

        # Define the expected return value
        expected_return = {
            'driver_volume_type': 'iscsi',
            'data': {
                'target_discovered': False,
                'target_portal': '192.168.0.2:3260',
                'target_lun': 0,
                'target_iqns': ['iqn.2023-06.com.toyou:initiator-group-uuid1'],
                'target_portals': ['192.168.0.2:3260'],
                'target_luns': [0]
            }
        }

        # Verify the method calls and return value
        self.mock_client.get_initiator_list.assert_called_once()
        self.mock_client.create_initiator_group.assert_called_once()
        self.assertEqual(
            self.mock_client.get_initiator_target_connections.call_count, 2)
        self.assertEqual(self.mock_client.get_target.call_count, 2)
        self.mock_client.modify_target.assert_not_called()
        self.mock_client.create_target.assert_called_once()
        self.mock_client.generate_config.assert_called_once()

        self.assertEqual(result, expected_return)

    def test_initialize_connection_failure(self):
        """Test case for failed volume initialization.

        - Sets the connector information.
        - Sets the mock return values for get_initiator_list and get_it.
        - Calls the initialize_connection method.
        - Verifies if the get_initiator_list method is called once.
        - Verifies if the create_initiator_group method is not called.
        - Verifies if the initialize_connection method raises an exception to
        type exception.VolumeBackendAPIException.
        """
        # Set the connector information
        connector = {
            'host': 'host1',
            'initiator': 'iqn.1234',
            'ip': '192.168.0.1',
            'uuid': 'uuid1'
        }

        # Set the mock return values for get_initiator_list and get_it
        self.mock_client.get_initiator_list.return_value = [
            {'group_name': 'initiator-group-uuid1'}]
        self.mock_client.get_initiator_target_connections.return_value = []

        # Call the initialize_connection method and check for exception
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection, self.volume,
                          connector)

        # Verify if the get_initiator_list method is called once
        self.mock_client.get_initiator_list.assert_called_once()

        # Verify if the create_initiator_group method is not called
        self.mock_client.create_initiator_group.assert_not_called()

    def test_terminate_connection_success(self):
        """Test case for successful termination of volume connection.

        - Sets the connector information.
        - Sets the mock return values for get_it and get_initiator_list.
        - Calls the terminate_connection method with the required mock methods.
        - Verifies the method calls using assertions.
        """
        # Set the connector information
        connector = {
            'host': 'host1',
            'initiator': 'iqn.1234',
            'ip': '192.168.0.1',
            'uuid': 'uuid1'
        }

        # Set the mock return values for get_it and get_initiator_list
        self.mock_client.get_initiator_target_connections.return_value = [
            {'target_iqn': 'target_iqn1',
             'target_name': 'target1',
             'hostName': ['host1'],
             'block': [{'name': 'volume1', 'lunid': 1},
                       {'name': 'volume2', 'lunid': 2}]}
        ]
        self.mock_client.get_initiator_list.return_value = [
            {'group_name': 'initiator-group-uuid1', 'group_id': 'group_id1'}
        ]

        # Call the terminate_connection method with the required mock methods
        self.driver.terminate_connection(
            self.volume,
            connector,
            mock_get_it=self.mock_client.get_initiator_target_connections,
            mock_delete_target=self.mock_client.delete_target,
            mock_get_initiator_list=self.mock_client.get_initiator_list,
            mock_delete_initiator_group=self.mock_client
            .delete_initiator_group,
            mock_restart_service=self.mock_client.restart_service,
        )

        # Verify the method calls using assertions
        self.mock_client.get_initiator_target_connections.assert_called_once()
        self.mock_client.get_initiator_list.assert_not_called()
        self.mock_client.delete_target.assert_not_called()
        self.mock_client.delete_initiator_group.assert_not_called()
        self.mock_client.restart_service.assert_not_called()

    def test_terminate_connection_failure(self):
        """Test case for failed termination of volume connection.

        - Sets the connector information.
        - Sets the mock return values for get_it and get_initiator_list.
        - Sets the delete_target method to raise an exception.
        - Calls the terminate_connection method.
        - Verifies the method calls and assertions.
        """
        # Set the connector information
        connector = {
            'host': 'host1',
            'initiator': 'iqn.1234',
            'ip': '192.168.0.1',
            'uuid': 'uuid1'
        }

        # Set the mock return values for get_it and get_initiator_list
        self.mock_client.get_initiator_target_connections.return_value = [
            {
                'target_iqn': 'target_iqn1',
                'target_name': 'iqn.2023-06.com.toyou:initiator-group-uuid1',
                'hostName': ['host1'],
                'block': [{'name': self.volume.name, 'lunid': 1}]
            }
        ]
        self.mock_client.get_initiator_list.return_value = [
            {'group_name': 'initiator-group-uuid1', 'group_id': 'group_id1'}
        ]

        # Set the delete_target method to raise an exception
        self.mock_client.delete_target.side_effect = \
            exception.VolumeBackendAPIException('API error')

        # Assert that an exception to type exception.VolumeBackendAPIException
        # is raised
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.terminate_connection,
                          self.volume,
                          connector)

        # Verify method calls and assertions
        self.mock_client.get_initiator_target_connections.assert_called_once()
        self.mock_client.get_initiator_list.assert_not_called()
        self.mock_client.delete_target.assert_called_once_with('target_iqn1')
        self.mock_client.delete_initiator_group.assert_not_called()
        self.mock_client.restart_service.assert_not_called()



# Copyright 2022 Red Hat, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest
from unittest import mock

import ddt

from cinder import exception
from cinder.tests.unit.privsep.targets import fake_nvmet_lib
from cinder.tests.unit import test
# This must go after fake_nvmet_lib has been imported (thus the noqa)
from cinder.privsep.targets import nvmet  # noqa


@ddt.ddt
class TestSerialize(test.TestCase):
    def setUp(self):
        super().setUp()
        fake_nvmet_lib.reset_mock()

    def test_tuple(self):
        """Test serialization of a tuple."""
        instance = (1, 'string')
        res = nvmet.serialize(instance)
        self.assertEqual(('tuple', instance), res)

    @ddt.data(1, 1.1, 'string', None, [1, 2, 'string'])
    def test_others(self, instance):
        """Test normal Python instances that should not be modified."""
        res = nvmet.serialize(instance)
        self.assertEqual(instance, res)

    def test_root(self):
        instance = nvmet.Root()
        res = nvmet.serialize(instance)
        self.assertEqual(('Root', {}), res)

    def test_host(self):
        instance = nvmet.Host(nqn='_nqn')
        res = nvmet.serialize(instance)
        self.assertEqual(('Host', {'nqn': '_nqn', 'mode': 'lookup'}), res)

    def test_subsystem(self):
        instance = nvmet.Subsystem(nqn='_nqn')
        res = nvmet.serialize(instance)
        self.assertEqual(('Subsystem', {'nqn': '_nqn', 'mode': 'lookup'}), res)

    def test_namespace(self):
        subsys = nvmet.Subsystem(nqn='_nqn')
        instance = nvmet.Namespace(subsystem=subsys, nsid='_nsid')
        res = nvmet.serialize(instance)
        # Subsystem is a recursive serialization
        expected = (
            'Namespace', {'subsystem': ('Subsystem', {'nqn': '_nqn',
                                                      'mode': 'lookup'}),
                          'nsid': '_nsid',
                          'mode': 'lookup'})
        self.assertEqual(expected, res)

    def test_port(self):
        instance = nvmet.Port(portid='_portid')
        res = nvmet.serialize(instance)
        expected = ('Port', {'portid': '_portid', 'mode': 'lookup'})
        self.assertEqual(expected, res)

    def test_Referral(self):
        port = nvmet.Port(portid='_portid')
        # name is a Mock attribute, so we'll use it as instance.name
        instance = nvmet.Referral(port=port, name='_name')
        res = nvmet.serialize(instance)
        # Port is a recursive serialization
        expected = (
            'Referral', {'port': ('Port', {'portid': '_portid',
                                           'mode': 'lookup'}),
                         'name': instance.name,
                         'mode': 'lookup'})
        self.assertEqual(expected, res)

    def test_ANAGroup(self):
        port = nvmet.Port(portid='_portid')
        instance = nvmet.ANAGroup(port=port, grpid='_grpid')
        res = nvmet.serialize(instance)
        expected = (
            'ANAGroup', {'port': ('Port', {'portid': '_portid',
                                           'mode': 'lookup'}),
                         'grpid': '_grpid',
                         'mode': 'lookup'})
        self.assertEqual(expected, res)


@ddt.ddt
class TestDeserialize(test.TestCase):
    def test_deserialize_tuple(self):
        """Test serialization of a tuple."""
        expected = (1, 'string')
        data = ('tuple', expected)
        res = nvmet.deserialize(data)
        self.assertEqual(expected, res)

    @ddt.data(1, 1.1, 'string', None, [1, 2, 'string'])
    def test_deserialize_others(self, data):
        """Test normal Python instances that should not be modified."""
        res = nvmet.deserialize(data)
        self.assertEqual(data, res)

    def test_deserialize_root(self):
        data = ('Root', {})
        res = nvmet.deserialize(data)
        self.assertIsInstance(res, nvmet.nvmet.Root)

    def test_deserialize_host(self):
        data = ('Host', {'nqn': '_nqn', 'mode': 'lookup'})
        host = nvmet.deserialize(data)
        self.assertIsInstance(host, nvmet.nvmet.Host)
        self.assertEqual('_nqn', host.nqn)
        self.assertEqual('lookup', host.mode)

    def test_deserialize_subsystem(self):
        data = ('Subsystem', {'nqn': '_nqn', 'mode': 'lookup'})
        subsys = nvmet.deserialize(data)
        self.assertIsInstance(subsys, nvmet.nvmet.Subsystem)
        self.assertEqual('_nqn', subsys.nqn)
        self.assertEqual('lookup', subsys.mode)

    def test_deserialize_namespace(self):
        data = ('Namespace', {'subsystem': ('Subsystem', {'nqn': '_nqn',
                                                          'mode': 'lookup'}),
                              'nsid': '_nsid',
                              'mode': 'lookup'})

        ns = nvmet.deserialize(data)
        self.assertIsInstance(ns, nvmet.nvmet.Namespace)
        self.assertEqual('_nsid', ns.nsid)
        self.assertEqual('lookup', ns.mode)
        self.assertIsInstance(ns.subsystem, nvmet.nvmet.Subsystem)
        self.assertEqual('_nqn', ns.subsystem.nqn)
        self.assertEqual('lookup', ns.subsystem.mode)

    def test_deserialize_port(self):
        data = ('Port', {'portid': '_portid', 'mode': 'lookup'})
        port = nvmet.deserialize(data)
        self.assertIsInstance(port, nvmet.nvmet.Port)
        self.assertEqual('_portid', port.portid)
        self.assertEqual('lookup', port.mode)

    def test_deserialize_Referral(self):
        data = ('Referral', {'port': ('Port', {'portid': '_portid',
                                               'mode': 'lookup'}),
                             'name': '1',
                             'mode': 'lookup'})
        ref = nvmet.deserialize(data)

        self.assertIsInstance(ref, nvmet.nvmet.Referral)
        self.assertEqual('1', ref._mock_name)  # Because name is used by Mock
        self.assertEqual('lookup', ref.mode)
        self.assertIsInstance(ref.port, nvmet.nvmet.Port)
        self.assertEqual('_portid', ref.port.portid)
        self.assertEqual('lookup', ref.port.mode)

    def test_deserialize_ANAGroup(self):
        data = ('ANAGroup', {'port': ('Port', {'portid': '_portid',
                                               'mode': 'lookup'}),
                             'grpid': '_grpid',
                             'mode': 'lookup'})
        ana = nvmet.deserialize(data)

        self.assertIsInstance(ana, nvmet.nvmet.ANAGroup)
        self.assertEqual('_grpid', ana.grpid)
        self.assertEqual('lookup', ana.mode)
        self.assertIsInstance(ana.port, nvmet.nvmet.Port)
        self.assertEqual('_portid', ana.port.portid)
        self.assertEqual('lookup', ana.port.mode)

    @mock.patch.object(nvmet, 'deserialize')
    def test_deserialize_params(self, mock_deserialize):
        mock_deserialize.side_effect = [11, 22, 33, 55, 77]
        args = [1, 2, 3]
        kwargs = {'4': 5, '6': 7}

        res_args, res_kwargs = nvmet.deserialize_params(args, kwargs)

        self.assertEqual(5, mock_deserialize.call_count)
        mock_deserialize.assert_has_calls((mock.call(1),
                                           mock.call(2),
                                           mock.call(3),
                                           mock.call(5),
                                           mock.call(7)))
        self.assertEqual([11, 22, 33], res_args)
        self.assertEqual({'4': 55, '6': 77}, res_kwargs)


class TestPrivsep(test.TestCase):
    @mock.patch.object(nvmet.LOG, 'error')
    def test__nvmet_setup_failure(self, mock_log):
        self.assertRaises(exception.CinderException,
                          nvmet._nvmet_setup_failure, mock.sentinel.message)
        mock_log.assert_called_once_with(mock.sentinel.message)

    @mock.patch.object(nvmet, '_privsep_setup')
    def test_privsep_setup(self, mock_setup):
        args = [mock.sentinel.arg1, mock.sentinel.arg2]
        kwargs = {'kwarg1': mock.sentinel.kwarg1}

        res = nvmet.privsep_setup('MyClass', err_func=None, *args, **kwargs)

        mock_setup.assert_called_once_with('MyClass', *args, **kwargs)
        self.assertEqual(mock_setup.return_value, res)

    @mock.patch.object(nvmet, '_privsep_setup')
    def test_privsep_setup_err_func_as_arg_none(self, mock_setup):
        exc = exception.CinderException('ouch')
        mock_setup.side_effect = exc
        args = [mock.sentinel.arg1, mock.sentinel.arg2, None]
        kwargs = {'kwarg1': mock.sentinel.kwarg1}

        # NOTE: testtools.TestCase were Cinder's tests inherit from masks the
        # unittest's assertRaises that supports context manager usage, so we
        # address it directly.
        with unittest.TestCase.assertRaises(self,
                                            exception.CinderException) as cm:
            nvmet.privsep_setup('MyClass', *args, **kwargs)

        self.assertEqual(exc, cm.exception)
        mock_setup.assert_called_once_with('MyClass', *args[:-1], **kwargs)

    @mock.patch.object(nvmet, '_privsep_setup')
    def test_privsep_setup_err_func_as_arg(self, mock_setup):
        def err_func(msg):
            raise exception.VolumeDriverException()

        mock_setup.side_effect = exception.CinderException('ouch')
        args = [mock.sentinel.arg1, mock.sentinel.arg2, err_func]

        self.assertRaises(exception.VolumeDriverException,
                          nvmet.privsep_setup, 'MyClass', *args)
        mock_setup.assert_called_once_with('MyClass', *args[:-1])

    # We mock the privsep context mode to fake that we are not the client
    @mock.patch('cinder.privsep.sys_admin_pctxt.client_mode', False)
    @mock.patch.object(nvmet, 'deserialize_params')
    @mock.patch.object(nvmet.nvmet, 'MyClass')
    def test__privsep_setup(self, mock_class, mock_deserialize):
        args = (1, 2, 3)
        kwargs = {'4': 5, '6': 7}
        deserialized_args = (11, 22, 33)
        deserialized_kwargs = {'4': 55, '6': 77}

        expected_args = deserialized_args[:]
        expected_kwargs = deserialized_kwargs.copy()
        expected_kwargs['err_func'] = nvmet._nvmet_setup_failure

        mock_deserialize.return_value = (deserialized_args,
                                         deserialized_kwargs)

        res = nvmet._privsep_setup('MyClass', *args, **kwargs)

        mock_deserialize.assert_called_once_with(args, kwargs)
        mock_class.setup.assert_called_once_with(*expected_args,
                                                 **expected_kwargs)
        self.assertEqual(mock_class.setup.return_value, res)

    # We mock the privsep context mode to fake that we are not the client
    @mock.patch('cinder.privsep.sys_admin_pctxt.client_mode', False)
    @mock.patch.object(nvmet, 'deserialize')
    @mock.patch.object(nvmet, 'deserialize_params')
    def test_do_privsep_call(self, mock_deserialize_params, mock_deserialize):
        args = (1, 2, 3)
        kwargs = {'4': 5, '6': 7}
        deserialized_args = (11, 22, 33)
        deserialized_kwargs = {'4': 55, '6': 77}

        mock_deserialize_params.return_value = (deserialized_args,
                                                deserialized_kwargs)

        res = nvmet.do_privsep_call(mock.sentinel.instance,
                                    'method_name',
                                    *args, **kwargs)
        mock_deserialize.assert_called_once_with(mock.sentinel.instance)
        mock_deserialize_params.assert_called_once_with(args, kwargs)

        mock_method = mock_deserialize.return_value.method_name
        mock_method.assert_called_once_with(*deserialized_args,
                                            **deserialized_kwargs)
        self.assertEqual(mock_method.return_value, res)


@ddt.ddt
class TestNvmetClasses(test.TestCase):
    @ddt.data('Host', 'Referral', 'ANAGroup')
    def test_same_classes(self, cls_name):
        self.assertEqual(getattr(nvmet, cls_name),
                         getattr(nvmet.nvmet, cls_name))

    def test_subsystem_init(self):
        subsys = nvmet.Subsystem('nqn')
        self.assertIsInstance(subsys, nvmet.nvmet.Subsystem)
        self.assertIsInstance(subsys, nvmet.Subsystem)
        self.assertEqual('nqn', subsys.nqn)
        self.assertEqual('lookup', subsys.mode)

    @mock.patch.object(nvmet, 'privsep_setup')
    def test_subsystem_setup(self, mock_setup):
        nvmet.Subsystem.setup(mock.sentinel.t, mock.sentinel.err_func)
        mock_setup.assert_called_once_with('Subsystem', mock.sentinel.t,
                                           mock.sentinel.err_func)

    @mock.patch.object(nvmet, 'privsep_setup')
    def test_subsystem_setup_no_err_func(self, mock_setup):
        nvmet.Subsystem.setup(mock.sentinel.t)
        mock_setup.assert_called_once_with('Subsystem', mock.sentinel.t, None)

    @mock.patch.object(nvmet, 'serialize')
    @mock.patch.object(nvmet, 'do_privsep_call')
    def test_subsystem_delete(self, mock_privsep, mock_serialize):
        subsys = nvmet.Subsystem('nqn')
        subsys.delete()
        mock_serialize.assert_called_once_with(subsys)
        mock_privsep.assert_called_once_with(mock_serialize.return_value,
                                             'delete')

    @mock.patch('os.listdir',
                return_value=['/path/namespaces/1', '/path/namespaces/2'])
    @mock.patch.object(nvmet, 'Namespace')
    def test_subsystem_namespaces(self, mock_nss, mock_listdir):
        subsys = nvmet.Subsystem(mock.sentinel.nqn)
        subsys.path = '/path'  # Set by the parent nvmet library Root class

        res = list(subsys.namespaces)

        self.assertEqual([mock_nss.return_value, mock_nss.return_value], res)

        mock_listdir.assert_called_once_with('/path/namespaces/')
        self.assertEqual(2, mock_nss.call_count)
        mock_nss.assert_has_calls((mock.call(subsys, '1'),
                                   mock.call(subsys, '2')))

    def test_port_init(self):
        port = nvmet.Port('portid')
        self.assertIsInstance(port, nvmet.nvmet.Port)
        self.assertIsInstance(port, nvmet.Port)
        self.assertEqual('portid', port.portid)
        self.assertEqual('lookup', port.mode)

    @mock.patch.object(nvmet, 'serialize')
    @mock.patch.object(nvmet, 'privsep_setup')
    def test_port_setup(self, mock_setup, mock_serialize):
        nvmet.Port.setup(mock.sentinel.root, mock.sentinel.n,
                         mock.sentinel.err_func)
        mock_serialize.assert_called_once_with(mock.sentinel.root)
        mock_setup.assert_called_once_with('Port', mock_serialize.return_value,
                                           mock.sentinel.n,
                                           mock.sentinel.err_func)

    @mock.patch.object(nvmet, 'serialize')
    @mock.patch.object(nvmet, 'privsep_setup')
    def test_port_setup_no_err_func(self, mock_setup, mock_serialize):
        nvmet.Port.setup(mock.sentinel.root, mock.sentinel.n)
        mock_serialize.assert_called_once_with(mock.sentinel.root)
        mock_setup.assert_called_once_with('Port', mock_serialize.return_value,
                                           mock.sentinel.n, None)

    @mock.patch.object(nvmet, 'serialize')
    @mock.patch.object(nvmet, 'do_privsep_call')
    def test_port_add_subsystem(self, mock_privsep, mock_serialize):
        port = nvmet.Port('portid')
        port.add_subsystem(mock.sentinel.nqn)
        mock_serialize.assert_called_once_with(port)
        mock_privsep.assert_called_once_with(mock_serialize.return_value,
                                             'add_subsystem',
                                             mock.sentinel.nqn)

    @mock.patch.object(nvmet, 'serialize')
    @mock.patch.object(nvmet, 'do_privsep_call')
    def test_port_remove_subsystem(self, mock_privsep, mock_serialize):
        port = nvmet.Port('portid')
        port.remove_subsystem(mock.sentinel.nqn)
        mock_serialize.assert_called_once_with(port)
        mock_privsep.assert_called_once_with(mock_serialize.return_value,
                                             'remove_subsystem',
                                             mock.sentinel.nqn)

    @mock.patch.object(nvmet, 'serialize')
    @mock.patch.object(nvmet, 'do_privsep_call')
    def test_port_delete(self, mock_privsep, mock_serialize):
        port = nvmet.Port('portid')
        port.delete()
        mock_serialize.assert_called_once_with(port)
        mock_privsep.assert_called_once_with(mock_serialize.return_value,
                                             'delete')

    @mock.patch('os.listdir', return_value=['/path/ports/1', '/path/ports/2'])
    @mock.patch.object(nvmet, 'Port')
    def test_root_ports(self, mock_port, mock_listdir):
        r = nvmet.Root()
        r.path = '/path'  # This is set by the parent nvmet library Root class

        res = list(r.ports)

        self.assertEqual([mock_port.return_value, mock_port.return_value], res)

        mock_listdir.assert_called_once_with('/path/ports/')
        self.assertEqual(2, mock_port.call_count)
        mock_port.assert_has_calls((mock.call('1'), mock.call('2')))

    def test_namespace_init(self):
        ns = nvmet.Namespace('subsystem', 'nsid')
        self.assertIsInstance(ns, nvmet.nvmet.Namespace)
        self.assertIsInstance(ns, nvmet.Namespace)
        self.assertEqual('subsystem', ns.subsystem)
        self.assertEqual('nsid', ns.nsid)
        self.assertEqual('lookup', ns.mode)

    @mock.patch.object(nvmet, 'serialize')
    @mock.patch.object(nvmet, 'privsep_setup')
    def test_namespace_setup(self, mock_setup, mock_serialize):
        nvmet.Namespace.setup(mock.sentinel.subsys,
                              mock.sentinel.n)
        mock_serialize.assert_called_once_with(mock.sentinel.subsys)
        mock_setup.assert_called_once_with('Namespace',
                                           mock_serialize.return_value,
                                           mock.sentinel.n, None)

    @mock.patch.object(nvmet, 'serialize')
    @mock.patch.object(nvmet, 'do_privsep_call')
    def test_namespace_delete(self, mock_privsep, mock_serialize):
        ns = nvmet.Namespace('subsystem', 'nsid')
        ns.delete()
        mock_serialize.assert_called_once_with(ns)
        mock_privsep.assert_called_once_with(mock_serialize.return_value,
                                             'delete')



# Copyright (c) 2019 FUJITSU LIMITED
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Cinder Volume driver for Fujitsu ETERNUS DX S3 series."""

from cinder.i18n import _
from cinder import ssh_utils
from cinder.volume.drivers.fujitsu.eternus_dx import constants as CONSTANTS


class FJDXCLI(object):
    """ETERNUS CLI Code."""

    def __init__(self, user, storage_ip, password=None, keyfile=None):
        """Constructor."""
        self.user = user
        self.storage_ip = storage_ip
        if password and keyfile:
            raise Exception(_('can not specify both password and keyfile'))

        self.use_ipv6 = False
        if storage_ip.find(':') != -1:
            self.use_ipv6 = True

        if password:
            self.ssh_pool = ssh_utils.SSHPool(storage_ip, 22, None, user,
                                              password=password, max_size=2)

        if keyfile:
            self.ssh_pool = ssh_utils.SSHPool(storage_ip, 22, None, user,
                                              privatekey=keyfile, max_size=2)

        self.ce_support = False
        self.CMD_dic = {
            'check_user_role': self._check_user_role,
            'expand_volume': self._expand_volume,
            'show_pool_provision': self._show_pool_provision,
            'show_qos_bandwidth_limit': self._show_qos_bandwidth_limit,
            'set_qos_bandwidth_limit': self._set_qos_bandwidth_limit,
            'set_volume_qos': self._set_volume_qos,
            'show_copy_sessions': self._show_copy_sessions,
            'show_volume_qos': self._show_volume_qos,
            'show_enclosure_status': self._show_enclosure_status,
            'start_copy_snap_opc': self._start_copy_snap_opc,
            'stop_copy_session': self._stop_copy_session,
            'start_copy_opc': self._start_copy_opc,
            'delete_volume': self._delete_volume
        }

    def done(self, command, **option):
        func = self.CMD_dic.get(command, self._default_func)
        return func(**option)

    def _exec_cli(self, cmd, StrictHostKeyChecking=True, **option):

        exec_cmdline = cmd + self._get_option(**option)
        stdoutdata = self._exec_cli_with_eternus(exec_cmdline)
        output = []
        message = []
        stdoutlist = stdoutdata.split('\r\n')
        output_header = ""

        for no, outline in enumerate(stdoutlist):
            if len(outline) <= 0 or outline is None:
                continue

            if not output_header.endswith(exec_cmdline):
                output_header += outline
                continue

            if 0 <= outline.find('Error'):
                raise Exception(_("Output: %(outline)s: "
                                  "Command: %(cmdline)s")
                                % {'outline': outline,
                                   'cmdline': exec_cmdline})

            if not self._is_status(outline):
                continue

            status = int(outline, 16)
            lineno = no + 1
            break
        else:
            raise Exception(_(
                "Invalid CLI output: %(exec_cmdline)s, %(stdoutlist)s")
                % {'exec_cmdline': exec_cmdline,
                   'stdoutlist': stdoutlist})

        if status == 0:
            rc = str(CONSTANTS.RC_OK)
            for outline in stdoutlist[lineno:]:
                if 0 <= outline.find('CLI>'):
                    continue
                if len(outline) <= 0:
                    continue
                if outline is None:
                    continue
                message.append(outline)
        else:
            code = stdoutlist[lineno]
            for outline in stdoutlist[lineno + 1:]:
                if 0 <= outline.find('CLI>'):
                    continue
                if len(outline) <= 0:
                    continue
                if outline is None:
                    continue
                output.append(outline)
            if cmd != "show cli-error-code":
                rc, message = self._create_error_message(code, output)
            else:
                rc = 'E' + code
                message = output

        return {'result': 0, 'rc': rc, 'message': message}

    def _exec_cli_with_eternus(self, exec_cmdline):
        """Execute CLI command with arguments."""
        ssh = None
        try:
            ssh = self.ssh_pool.get()
            chan = ssh.invoke_shell()
            chan.send(exec_cmdline + '\n')
            stdoutdata = ''
            while True:
                temp = chan.recv(65535)
                if isinstance(temp, bytes):
                    temp = temp.decode('utf-8')
                else:
                    temp = str(temp)
                stdoutdata += temp

                # CLI command end with 'CLI>'.
                if stdoutdata == '\r\nCLI> ':
                    continue
                if (stdoutdata[len(stdoutdata) - 5: len(stdoutdata) - 1] ==
                        'CLI>'):
                    break
        except Exception as e:
            raise Exception(_("Execute CLI "
                              "command error. Error: %s") % e)
        finally:
            if ssh:
                self.ssh_pool.put(ssh)
                self.ssh_pool.remove(ssh)
        return stdoutdata

    def _show_cli_error_message(self, **option):
        """Get error messages by error code."""
        output = self._exec_cli("show cli-error-code",
                                **option)
        rc = output['rc']
        if rc != str(CONSTANTS.RC_OK):
            raise Exception(_('_show_cli_error_message failed. '
                              'Return code: %lu') % rc)
        message = output['message'][1]
        output['message'] = message.split('\t')[1]
        return output

    def _create_error_message(self, code, msg):
        """Create error code and message using arguements."""
        message = None
        rc = 'E' + code
        try:
            option = {
                'error-code': code
            }
            message = self._show_cli_error_message(**option)['message']
        except Exception:
            message = CONSTANTS.CLIRETCODE_dic.get(rc, msg)

        return rc, message

    @staticmethod
    def _is_status(value):
        """Check whether input value is status value or not."""
        try:
            if len(value) != 2:
                return False

            int(value, 16)
            int(value[0], 16)
            int(value[1], 16)

            return True
        except ValueError:
            return False

    @staticmethod
    def _get_option(**option):
        """Create option strings from dictionary."""
        ret = ""
        for key, value in option.items():
            ret += " -%(key)s %(value)s" % {'key': key, 'value': value}
        return ret

    def _default_func(self, **option):
        """Default function."""
        raise Exception(_("Invalid function is specified"))

    def _check_user_role(self, **option):
        """Check user role."""
        try:
            output = self._exec_cli("show users",
                                    StrictHostKeyChecking=False,
                                    **option)
            # Return error.
            rc = output['rc']
            if rc != str(CONSTANTS.RC_OK):
                return output

            userlist = output.get('message')
            role = None
            for userinfo in userlist:
                username = userinfo.split('\t')[0]
                if username == self.user:
                    role = userinfo.split('\t')[1]
                    break

            output['message'] = role
        except Exception as ex:
            if 'show users' in str(ex):
                msg = ("Specified user(%s) does not have Software role"
                       % self.user)
            elif 'Error connecting' in str(ex):
                msg = (str(ex)[34:] +
                       ', Please check fujitsu_private_key_path or .xml file')
            else:
                msg = str(ex)
            output = {
                'result': 0,
                'rc': str(CONSTANTS.RC_FAILED),
                'message': msg
            }
        return output

    def _expand_volume(self, **option):
        """Exec expand volume."""
        return self._exec_cli("expand volume", **option)

    def _set_volume_qos(self, **option):
        """Exec set volume-qos."""
        return self._exec_cli("set volume-qos", **option)

    def _show_pool_provision(self, **option):
        """Get TPP provision capacity information."""
        try:
            output = self._exec_cli("show volumes", **option)

            rc = output['rc']

            if rc != str(CONSTANTS.RC_OK):
                return output

            clidatalist = output.get('message')

            data = 0
            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                if clidata[0] == 'FFFF':
                    break
                data += int(clidata[7], 16)
            provision = data / 2048

            output['message'] = provision
        except Exception as ex:
            output = {
                'result': 0,
                'rc': str(CONSTANTS.RC_FAILED),
                'message': "show pool provision capacity error: %s" % ex
            }

        return output

    def _show_copy_sessions(self, **option):
        """Get copy sessions."""
        try:
            output = self._exec_cli("show copy-sessions", **option)

            # return error
            rc = output['rc']

            if rc != str(CONSTANTS.RC_OK):
                return output

            cpsdatalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                # Get CopyType
                if clidata[2] == '01':
                    # CopyKind: OPC
                    if bin(int(clidata[3], 16) & 16) != 0:
                        # eg. 0b10010000
                        temp_type = 'Snap'
                    elif bin(int(clidata[3], 16) & 64) != 0:
                        # eg. 0b11000000
                        temp_type = 'Snap+'
                    else:
                        temp_type = 'Other'
                elif clidata[2] == '02':
                    # CopyKind: EC
                    if clidata[5] == 'FF':
                        temp_type = 'EC'
                    elif clidata[5] == '10':
                        temp_type = 'Sync_REC'
                    else:
                        temp_type = 'Other'
                else:
                    temp_type = 'Other'

                # Get Phases
                if clidata[6] == '00':
                    temp_phase = 'No_Pair'
                elif clidata[6] == '01':
                    temp_phase = 'Copying'
                elif clidata[6] == '02':
                    temp_phase = 'Equivalent'
                elif clidata[6] == '03':
                    temp_phase = 'Tracking'
                elif clidata[6] == '04':
                    temp_phase = 'Tracking_Copying'
                elif clidata[6] == '06':
                    temp_phase = 'Readying'
                else:
                    temp_phase = 'Other'

                # Get CopyStatus
                if clidata[7] == '00':
                    temp_status = 'Idle'
                elif clidata[7] == '01':
                    temp_status = 'Reserve'
                elif clidata[7] == '02':
                    temp_status = 'Active'
                elif clidata[7] == '03':
                    temp_status = 'Error_Suspend'
                elif clidata[7] == '04':
                    temp_status = 'Suspend'
                elif clidata[7] == '05':
                    temp_status = 'Halt'
                else:
                    temp_status = 'Other'

                cpsdatalist.append({'Source Num': int(clidata[13], 16),
                                    'Dest Num': int(clidata[14], 16),
                                    'Type': temp_type,
                                    'Status': temp_status,
                                    'Phase': temp_phase,
                                    'Session ID': int(clidata[0], 16)})

            output['message'] = cpsdatalist
        except Exception as ex:
            output = {'result': 0,
                      'rc': str(CONSTANTS.RC_FAILED),
                      'message': "Show copy sessions error: %s"
                                 % str(ex)}

        return output

    def _show_qos_bandwidth_limit(self, **option):
        """Get qos bandwidth limit."""
        clidata = None
        try:
            output = self._exec_cli("show qos-bandwidth-limit", **option)

            # return error
            rc = output['rc']

            if rc != str(CONSTANTS.RC_OK):
                return output

            qoslist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                qoslist.append({'total_limit': int(clidata[0], 16),
                                'total_iops_sec': int(clidata[1], 16),
                                'total_bytes_sec': int(clidata[2], 16),
                                'read_limit': int(clidata[0], 16),
                                'read_iops_sec': int(clidata[3], 16),
                                'read_bytes_sec': int(clidata[4], 16),
                                'write_limit': int(clidata[0], 16),
                                'write_iops_sec': int(clidata[5], 16),
                                'write_bytes_sec': int(clidata[6], 16)})

            output['message'] = qoslist

        except IndexError as ex:
            msg = ('The results returned by cli are not as expected. '
                   'Exception string: %s' % clidata)
            output = {'result': 0,
                      'rc': str(CONSTANTS.RC_FAILED),
                      'message': "Show qos bandwidth limit error: %s. %s"
                                 % (ex, msg)}

        except Exception as ex:
            output = {'result': 0,
                      'rc': str(CONSTANTS.RC_FAILED),
                      'message': "Show qos bandwidth limit error: %s" % ex}

        return output

    def _set_qos_bandwidth_limit(self, **option):
        """Set qos bandwidth limit"""
        return self._exec_cli("set qos-bandwidth-limit", **option)

    def _show_volume_qos(self, **option):
        """Get volumes with qos."""
        clidata = None
        try:
            output = self._exec_cli("show volume-qos", **option)

            # return error
            rc = output['rc']

            if rc != str(CONSTANTS.RC_OK):
                return output

            vqosdatalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                vqosdatalist.append({'total_limit': int(clidata[2], 16),
                                     'read_limit': int(clidata[3], 16),
                                     'write_limit': int(clidata[4], 16)})

            output['message'] = vqosdatalist

        except IndexError as ex:
            msg = ('The results returned by cli are not as expected. '
                   'Exception string: %s' % clidata)
            output = {'result': 0,
                      'rc': str(CONSTANTS.RC_FAILED),
                      'message': "Show volume qos error: %s. %s" % (ex, msg)}

        except Exception as ex:
            output = {'result': 0,
                      'rc': str(CONSTANTS.RC_FAILED),
                      'message': "Show volume qos error: %s" % ex}

        return output

    def _show_enclosure_status(self, **option):
        """Get the version of machine."""
        clidata = None
        try:
            output = self._exec_cli("show enclosure-status", **option)

            # return error
            rc = output['rc']

            if rc != str(CONSTANTS.RC_OK):
                return output

            clidatalist = output.get('message')
            clidata = clidatalist[0].split('\t')
            versioninfo = {'version': clidata[11]}

            output['message'] = versioninfo

        except IndexError as ex:
            msg = ('The results returned by cli are not as expected. '
                   'Exception string: %s' % clidata)
            output = {'result': 0,
                      'rc': str(CONSTANTS.RC_FAILED),
                      'message': "Show enclosure status error: %s. %s"
                                 % (ex, msg)}

        except Exception as ex:
            output = {'result': 0,
                      'rc': str(CONSTANTS.RC_FAILED),
                      'message': "Show enclosure status error: %s" % ex}

        return output

    def _start_copy_snap_opc(self, **option):
        """Exec start copy-snap-opc."""
        return self._exec_cli("start copy-snap-opc", **option)

    def _stop_copy_session(self, **option):
        """Exec stop copy-session."""
        return self._exec_cli("stop copy-session", **option)

    def _start_copy_opc(self, **option):
        """Exec start copy-opc."""
        return self._exec_cli("start copy-opc", **option)

    def _delete_volume(self, **option):
        """Exec delete volume."""
        return self._exec_cli('delete volume', **option)



#    (c)  Copyright 2022 Fungible, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from cinder import exception
from cinder.volume.drivers.fungible import constants
from cinder.volume.drivers.fungible import \
    swagger_api_client as swagger_client

LOG = logging.getLogger(__name__)


class RestClient(object):
    def __init__(self, configuration):
        """Initialize the api request fields."""
        self.configuration = configuration
        self.rest_ip = None
        self.rest_port = None
        self.is_configured = False

    @staticmethod
    def log_error(error_msg):
        """Raise exception with error message"""
        LOG.exception(error_msg)
        raise exception.VolumeBackendAPIException(data=error_msg)

    def do_setup(self):
        """Initial setup of API request variables"""
        get_config_value = self.configuration.safe_get
        self.client = swagger_client.Configuration()
        self.client.username = get_config_value("san_login")
        self.client.password = get_config_value("san_password")
        self.rest_ip = get_config_value("san_ip")
        self.rest_port = get_config_value("san_api_port")
        protocol = "https"
        self.client.host = f"{protocol}://{self.rest_ip}{constants.STATIC_URL}"
        self.client.verify_ssl = False
        if not self.configuration.api_enable_ssl:
            protocol = "http"
            self.client.host = (f"{protocol}://{self.rest_ip}:{self.rest_port}"
                                f"{constants.STATIC_URL}")
        LOG.info("REST server IP: %(ip)s, port: %(port)s, "
                 "username: %(user)s.",
                 {
                     "ip": self.rest_ip,
                     "port": self.rest_port,
                     "user": self.client.username,
                 })
        self.api_storage = swagger_client.StorageApi(
            swagger_client.ApiClient(self.client))
        self.api_gateway = swagger_client.ApigatewayApi(
            swagger_client.ApiClient(self.client))
        self.api_topology = swagger_client.TopologyApi(
            swagger_client.ApiClient(self.client))
        self.is_configured = True

    def check_for_setup_error(self):
        """Check status of fungible storage clusters."""
        api_response = self.api_gateway.get_fc_health().to_dict()
        return api_response

    def create_volume(self, volume, fungible_specs, volume_type,
                      snapshot=None):
        """Creates new volume using the specified parameters"""

        # Convert GB to bytes, Default 1 GB size
        volume_size = constants.BYTES_PER_GIB
        if volume['size']:
            volume_size = constants.BYTES_PER_GIB * volume['size']
        fungible_request_obj = {
            "name": volume['id'],
            "vol_type": volume_type.upper(),
            "capacity": volume_size,
            "is_clone": False,
            "encrypt": False,
            "qos_band": constants.QOS_BAND.get('silver'),
            "block_size": int(constants.BLOCK_SIZE_4K)
        }
        data_protection = {
            "num_failed_disks": 2,
            "num_data_disks": 4,
            "num_redundant_dpus": 1,
        }
        durable_param = {
            "compression_effort": 2,
            "snap_support": True,
            "space_allocation_policy": 'balanced',
        }
        # Create Volume From Snapshot
        if snapshot is not None:
            fungible_request_obj["is_clone"] = True
            fungible_request_obj["clone_source_volume_uuid"] = \
                snapshot['provider_id']

        errors = []
        # Validation check for Extraspecs
        self._validation_check(fungible_request_obj,
                               fungible_specs, volume_type, errors,
                               "ExtraSpecs", durable_param, data_protection)
        # Validation check for Metadata
        self._validation_check(fungible_request_obj,
                               volume.get('metadata', {}), volume_type, errors,
                               "Metadata", durable_param, data_protection)
        if len(errors) != 0:
            msg = "ERROR: "
            for error in errors:
                msg = msg + " | " + error
            self.log_error(error_msg=msg)

        LOG.info("create_volume: "
                 "fungible_request_obj=%(fungible_request_obj)s",
                 {'fungible_request_obj': fungible_request_obj})
        api_response = self.api_storage.create_volume(
            body_volume_intent_create=fungible_request_obj).to_dict()

        return api_response

    def _validation_check(self, fungible_obj, data, volume_type, errors,
                          prefix, durable_param, data_protection):
        if constants.FSC_KMIP_SECRET_KEY in data:
            if data[constants.FSC_KMIP_SECRET_KEY]:
                fungible_obj['encrypt'] = True
                fungible_obj['kmip_secret_key'] = data[
                    constants.FSC_KMIP_SECRET_KEY]

        if constants.FSC_BLK_SIZE in data:
            if data[constants.FSC_BLK_SIZE] in constants.BLOCK_SIZE:
                if (volume_type.upper() == constants.VOLUME_TYPE_RF1 and
                        data[constants.FSC_BLK_SIZE] !=
                        constants.BLOCK_SIZE_16K):
                    msg = (
                        f"{prefix} {constants.FSC_BLK_SIZE} value is invalid \
                        for the volume type specified")
                    errors.append(msg)
                else:
                    fungible_obj['block_size'] = int(
                        data[constants.FSC_BLK_SIZE])
            else:
                msg = (f"{prefix} {constants.FSC_BLK_SIZE} value is invalid")
                errors.append(msg)
        elif volume_type.upper() == constants.VOLUME_TYPE_RF1:
            # Set default block size for RF1 to 16K
            fungible_obj['block_size'] = int(constants.BLOCK_SIZE_16K)

        if constants.FSC_QOS_BAND in data:
            if data[constants.FSC_QOS_BAND].lower() in constants.QOS_BAND:
                fungible_obj['qos_band'] = constants.QOS_BAND.get(
                    data[constants.FSC_QOS_BAND].lower())
            else:
                msg = (f"{prefix} {constants.FSC_QOS_BAND} value is invalid")
                errors.append(msg)

        if (volume_type.upper() == constants.VOLUME_TYPE_RAW or
                volume_type.upper() == constants.VOLUME_TYPE_RF1):
            if constants.FSC_FD_IDS in data:
                ids = data[constants.FSC_FD_IDS].split(',', 2)
                if len(ids) <= 2:
                    ids = [item.strip() for item in ids]
                    fungible_obj['fault_domain_ids'] = ids
                else:
                    msg = (f"{prefix} {constants.FSC_FD_IDS} - "
                           f"Only two fault domain ids can be provided.")
                    errors.append(msg)

            if constants.FSC_FD_OP in data:
                if (data[constants.FSC_FD_OP].upper()
                        in constants.FSC_FD_OPS):
                    fungible_obj['fd_op'] = data[constants.FSC_FD_OP]
                else:
                    msg = (f"{prefix} {constants.FSC_FD_OP} "
                           f"value is invalid")
                    errors.append(msg)

        if (volume_type.upper() == constants.VOLUME_TYPE_REPLICA or
                volume_type.upper() == constants.VOLUME_TYPE_EC or
                volume_type.upper() == constants.VOLUME_TYPE_RF1):
            if constants.FSC_SPACE_ALLOCATION_POLICY in data:
                if (data[constants.FSC_SPACE_ALLOCATION_POLICY].lower()
                        in constants.SPACE_ALLOCATION_POLICY):
                    durable_param['space_allocation_policy'] = data[
                        constants.FSC_SPACE_ALLOCATION_POLICY]
                else:
                    msg = (f"{prefix} {constants.FSC_SPACE_ALLOCATION_POLICY}"
                           f" value is invalid")
                    errors.append(msg)

            if constants.FSC_COMPRESSION in data:
                if (data[constants.FSC_COMPRESSION].lower() ==
                        constants.FALSE):
                    durable_param['compression_effort'] = 0
                elif (data[constants.FSC_COMPRESSION].lower() ==
                      constants.TRUE):
                    durable_param['compression_effort'] = 2
                else:
                    msg = (f"{prefix} {constants.FSC_COMPRESSION} value is "
                           f"invalid")
                    errors.append(msg)

            if constants.FSC_SNAPSHOTS in data:
                if (data[constants.FSC_SNAPSHOTS].lower()
                        in constants.BOOLEAN):
                    if (data[constants.FSC_SNAPSHOTS].lower() ==
                            constants.FALSE):
                        durable_param['snap_support'] = False
                else:
                    msg = (f"{prefix} {constants.FSC_SNAPSHOTS} value is "
                           f"invalid")
                    errors.append(msg)

        if volume_type.upper() == constants.VOLUME_TYPE_EC:
            fungible_obj.update(durable_param)
            if constants.FSC_EC_SCHEME in data:
                if data[constants.FSC_EC_SCHEME] == constants.EC_8_2:
                    data_protection['num_data_disks'] = 8
                elif data[constants.FSC_EC_SCHEME] == constants.EC_4_2:
                    data_protection['num_data_disks'] = 4
                elif data[constants.FSC_EC_SCHEME] == constants.EC_2_1:
                    data_protection['num_data_disks'] = 2
                    data_protection['num_failed_disks'] = 1
                else:
                    msg = (f"{prefix} {constants.FSC_EC_SCHEME} value is "
                           f"invalid")
                    errors.append(msg)
            fungible_obj["data_protection"] = data_protection

        elif volume_type.upper() == constants.VOLUME_TYPE_REPLICA:
            fungible_obj.update(durable_param)
            data_protection = {
                "num_failed_disks": 1,
                "num_data_disks": 1,
                "num_redundant_dpus": 1,
            }
            fungible_obj["data_protection"] = data_protection

        elif volume_type.upper() == constants.VOLUME_TYPE_RF1:
            fungible_obj.update(durable_param)
            pass

    def delete_volume(self, volume_uuid):
        """Deletes the specified volume"""
        LOG.info("delete_volume: volume_uuid=%(volume_uuid)s",
                 {'volume_uuid': volume_uuid})
        api_response = self.api_storage.delete_volume(
            volume_uuid=volume_uuid).to_dict()

        return api_response

    def get_volume_detail(self, uuid):
        """Get volume details by uuid"""
        api_response = self.api_storage.get_volume(volume_uuid=uuid).to_dict()

        return api_response

    def get_host_uuid_from_host_nqn(self, host_nqn):
        """Get host uuid from the host_nqn supplied"""
        api_response = self.api_topology.get_host_id_list(
            host_nqn_contains=host_nqn).to_dict()

        host_uuids = api_response.get("data").get("host_uuids")
        if len(host_uuids) == 1:
            return host_uuids[0]
        else:
            return None

    def get_host_details(self, host_uuid):
        """Get host details for the host_uuid supplied"""
        api_response = self.api_topology.get_host_info(
            host_uuid=host_uuid).to_dict()

        host = api_response.get("data")
        return host

    def get_hosts_subset(self, host_uuids):
        """Get host details in a list for the list of host_uuids supplied"""
        request_obj = {
            "host_id_list": host_uuids
        }
        api_response = self.api_topology.fetch_hosts_with_ids(
            body_fetch_hosts_with_ids=request_obj).to_dict()

        hosts = api_response.get("data")
        return hosts

    def create_host(self, host_nqn):
        """Create host with the host_nqn supplied"""
        request_obj = {
            "host_name": host_nqn,
            "host_nqn": host_nqn,
            "fac_enabled": False
        }
        LOG.info("create_host: request_obj=%(request_obj)s",
                 {'request_obj': request_obj})
        api_response = self.api_topology.add_host(
            body_host_create=request_obj).to_dict()

        return api_response

    def attach_volume(self, uuid, host_uuid, fac_enabled, iops=False):
        """Attaches a volume to a host server,

        using the specified transport method
        """
        if fac_enabled:
            request_obj = {
                "transport": 'PCI',
                "host_uuid": host_uuid,
                "fnid": 3,
                "huid": 1,
                "ctlid": 0
            }
        else:
            request_obj = {
                "transport": 'TCP',
                "host_uuid": host_uuid
            }
        # high iops set when uploading, downloading or migrating volume
        if iops:
            request_obj["max_read_iops"] = self.configuration.safe_get(
                'iops_for_image_migration')

        LOG.info("attach_volume: uuid=%(uuid)s "
                 "request_obj=%(request_obj)s",
                 {'uuid': uuid, 'request_obj': request_obj})
        api_response = self.api_storage.attach_volume(
            volume_uuid=uuid,
            body_volume_attach=request_obj).to_dict()

        return api_response

    def detach_volume(self, port_uuid):
        """Detach the volume specified port"""
        LOG.info("detach_volume: port_uuid=%(port_uuid)s",
                 {'port_uuid': port_uuid})
        api_response = self.api_storage.delete_port(
            port_uuid=port_uuid).to_dict()

        return api_response

    def create_snapshot(self, uuid, snapshot_name):
        """Create snapshot of volume with specified uuid"""
        fungible_request_obj = {
            "name": snapshot_name
        }
        api_response = self.api_storage.create_snapshot(
            volume_uuid=uuid,
            body_volume_snapshot_create=fungible_request_obj).to_dict()

        return api_response

    def delete_snapshot(self, uuid):
        """Delete snapshot with specified uuid"""
        api_response = self.api_storage.delete_snapshot(
            snapshot_uuid=uuid).to_dict()

        return api_response

    def extend_volume(self, uuid, new_size):
        """Update volume size to new size"""
        fungible_request_obj = {
            "op": "UPDATE_CAPACITY",
            "capacity": constants.BYTES_PER_GIB * new_size,
        }
        api_response = self.api_storage.update_volume(
            volume_uuid=uuid,
            body_volume_update=fungible_request_obj).to_dict()

        return api_response

    def rename_volume(self, uuid, new_name):
        """Update volume name to new name"""
        fungible_request_obj = {
            "op": "RENAME_VOLUME",
            "new_vol_name": new_name,
        }
        api_response = self.api_storage.update_volume(
            volume_uuid=uuid,
            body_volume_update=fungible_request_obj).to_dict()

        return api_response

    def copy_volume(self, volumeId, src_vrefId):
        """Submit copy volume task."""
        payload = {
            "src_volume_uuid": src_vrefId,
            "dest_volume_uuid": volumeId,
            "timeout": self.configuration.safe_get(
                'fsc_clone_volume_timeout')
        }

        LOG.info("Volume clone payload: %(payload)s.", {'payload': payload})
        api_response = self.api_storage.create_volume_copy_task(
            body_create_volume_copy_task=payload).to_dict()

        return api_response

    def get_volume_copy_task(self, task_uuid):
        """Get volume copy task status"""
        api_response = self.api_storage.get_volume_copy_task(
            task_uuid).to_dict()

        return api_response

    def delete_volume_copy_task(self, task_uuid):
        """Delete volume copy task"""
        api_response = self.api_storage.delete_volume_copy_task(
            task_uuid).to_dict()

        return api_response

    def get_topology(self):
        api_response = self.api_topology.get_hierarchical_topology().to_dict()
        return api_response
